# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]

__title__ = "acitoolkit"
__summary__ = "Toolkit for Cisco ACI Fabrics"
__uri__ = "http://datacenter.github.io/acitoolkit/"

__version__ = "0.3.1"

__author__ = "Cisco Systems, Inc."
__email__ = "acitoolkit@external.cisco.com"

__license__ = "Apache License, Version 2.0"
__copyright__ = "2015 %s" % __author__
